#include <stdio.h>
#include <cm/cmmd.h>

#if CMMD_VERSION == 20
#include <cm/cmmd-io.h>
#endif

#include "matrix.h"

#define  MATRIX_SIZE   352

int grid_to_global(int P, int Q, int p, int q)
{
   return(p * P + q);
}

void distribute(matrix *Mat, int M, int N)
{
   int P;
   int Q;
   int p;
   int q;

   int my_address;

   Mat -> num_nodes = CMMD_partition_size();
   Mat -> my_address = my_address = CMMD_self_address();

   Mat -> P = P = 4;
   Mat -> Q = Q = 4;

   Mat -> p = p = my_address / P;
   Mat -> q = q = my_address % P;

   Mat -> M = M;
   Mat -> N = N;

   Mat -> m = (M / P) + (p < (M % P));
   Mat -> n = (N / Q) + (q < (N % Q));
#if 0
   printf("%d: (p, q) = (%d, %d), (m, n) = (%d, %d)\n", my_address, p, q,
          Mat -> m, Mat -> n);
#endif
}

void make_matrix(matrix *Mat, int M, int N)
{
   double **Matrix;
   double  *Store;
   int      i;
   int      Index;
   int      m;
   int      n;

   distribute(Mat, M, N);

   m = Mat -> m;
   n = Mat -> n;

   if ((Matrix = (double **) malloc(m * sizeof(double *))) != NULL)
   {
      if ((Store = (double *) malloc(m * n * sizeof(double))) != NULL)
      {
         for (i = 0, Index = 0; i < m; i++, Index += n)
            Matrix[i] = &(Store[Index]);
      }
      else
      {
         printf("Out of memory!\n");
         exit(1);
      }
   }
   else
   {
      printf("Out of memory!\n");
      exit(1);
   }

   Mat -> Matrix = Matrix;
}

void fill_matrix(matrix *Mat)
{
   int      m = Mat -> m;
   int      n = Mat -> n;
   double **Matrix = Mat -> Matrix;
   int      i;
   int      j;

#if 1
   for (i = 0; i < m; i++)
      for (j = 0; j < n; j++)
         Matrix[i][j] = random();
#endif

#if 0
   if (Mat -> p == Mat -> q)
   {
      for (i = 0; i < m; i++)
         Matrix[i][i] = 1.0;
   }
#endif

#if 0
   for (i = 0; i < m; i++)
      for (j = 0; j < n; j++)
         Matrix[i][j] = 1.0;
#endif
}

void clear_matrix(matrix *Mat)
{
   int      m = Mat -> m;
   int      n = Mat -> n;
   double **Matrix = Mat -> Matrix;
   int      i;
   int      j;

   for (i = 0; i < m; i++)
      for (j = 0; j < n; j++)
         Matrix[i][j] = 0.0;
}

void mult_local_matrix(int m, int n, double **a, double **b, double **c)
{
   int i;
   int j;
   int k;

   for (i = 0; i < m; i++)
      for (k = 0; k < m; k++)
         for (j = 0; j < n; j++)
            c[i][k] += a[i][j] * b[j][k];
}

void send_handler(CMMD_mcb *mcb, void *arg)
{
   CMMD_free_mcb(*mcb);
}

void receive_handler(CMMD_mcb *mcb, int *Status)
{
   *Status = CMMD_mcb_tag(*mcb);
   CMMD_free_mcb(*mcb);
}

void matrix_mult(matrix *A, matrix *B, matrix *C, matrix *temp_Storage1,
                 matrix *temp_Storage2, matrix *temp_Storage3)
{
   int      q = A -> q;
   int      Q = A -> Q;
   int      p = A -> p;
   int      P = A -> P;
   int      m = A -> m;
   int      n = A -> n;
   int      k;
   int      j;
   int      qhat;
   int      phat;

   int      Boardcast_src;
   int      Boardcast_dest;
   int      Boardcast_tag;
   CMMD_mcb Boardcast_mcb;

   int      Buffer_Size;

   int      Shift_dest;
   int      Shift_src;
   int      Shift_tag;
   CMMD_mcb Shift_mcb;

   double **a = A -> Matrix;
   double **b = B -> Matrix;
   double **c = C -> Matrix;
   double **temp_a1 = temp_Storage1 -> Matrix;
   double **temp_a2 = temp_Storage2 -> Matrix;
   int      free_temp_a;

   double **working_a;
   double **Buffer_a;
   int      working_a_status;

   double **temp_b1 = b;
   double **temp_b2 = temp_Storage3 -> Matrix;
   int      free_temp_b;

   double **working_b;
   double **Buffer_b;
   int      working_b_status;

   qhat = phat = A -> p;
   Buffer_Size = m * n * sizeof(double);

   Boardcast_tag = A_TAG;

   if (q == qhat)
   {
      /* boardcast the local matrix  */
      for (j = 1; j < Q; j++)
      {
#if 0
         printf("k = 0: (%d, %d) -- sending out local A matrix to (%d, %d).\n",
                p, q, p, (qhat + j) % Q);
#endif
         Boardcast_dest = grid_to_global(P, Q, p, (qhat + j) % Q);
         CMMD_send_async(Boardcast_dest, Boardcast_tag, a[0], Buffer_Size,
                         send_handler, NULL); 
      }
      working_a_status = Boardcast_tag;
      free_temp_a = 1;
   }
   else
   {
      /* boardcast the local matrix  */
#if 0
      printf("k = 0: (%d, %d) -- receiving local A matrix from (%d, %d).\n",
             p, q, p, qhat);
#endif
      working_a_status = NOT_READY;
      Boardcast_src = grid_to_global(P, Q, p, qhat);
      Boardcast_mcb = CMMD_receive_async(Boardcast_src, Boardcast_tag,
                                         temp_a1[0], Buffer_Size,
                                         receive_handler, &working_a_status); 
      free_temp_a = 2;
   }

#if 0
   printf("shift addresss:\n");
#endif
   phat = p - 1;
   if (phat < 0)
      phat = P - 1;
#if 0
   printf("my address = (%d, %d), send to = (%d, %d)\n", p, q, phat, q);
#endif
   Shift_dest = grid_to_global(P, Q, phat, q);

   phat = (p + 1) % P;
#if 0
   printf("my address = (%d, %d), receive from = (%d, %d)\n", p, q, phat, q);
#endif
   Shift_src = grid_to_global(P, Q, phat, q);

   free_temp_b = 2;
   working_b = temp_b1;
   Buffer_b = temp_b2;
   working_b_status = Shift_tag = B_TAG;

   k = 0;
   do {

      if (q == qhat)
      {
         working_a = a;
         Buffer_a = temp_a1;
         free_temp_a = 2;
      }
      else
      {
         if (free_temp_a == 2)
         {
            working_a = temp_a1;
            Buffer_a = temp_a2;
         }
         else
         {
            working_a = temp_a2;
            Buffer_a = temp_a1;
         }
      }


      if (free_temp_b == 1)
      {
         working_b = temp_b2;
         Buffer_b = temp_b1;
         free_temp_b = 2;
      }
      else
      {
         working_b = temp_b1;
         Buffer_b = temp_b2;
         free_temp_b = 1;
      }

      k++;

      if (working_b_status != Shift_tag)
      {
#if 1
         printf("%d: (%d, %d) waiting for b\n", k, p, q);
#endif
         CMMD_msg_wait(Shift_mcb);
      }

      if (working_a_status != Boardcast_tag)
      {
#if 1
         printf("%d: (%d, %d) waiting for a\n", k, p, q);
#endif
         CMMD_msg_wait(Boardcast_mcb);
      }

      /* Send out the data for next stage */
      if (k < P)
      {
         Shift_tag = B_TAG + k;
         CMMD_send_async(Shift_dest, Shift_tag, working_b[0], Buffer_Size,
                         send_handler, NULL);

         Shift_mcb = CMMD_receive_async(Shift_src, Shift_tag, Buffer_b[0],
                                        Buffer_Size, receive_handler,
                                        &working_b_status); 

         qhat = (qhat + 1) % Q;
         Boardcast_tag = A_TAG + k;

         if (q == qhat)
         {
            /* boardcast the local matrix  */
            for (j = 1; j < Q; j++)
            {
#if 0
               printf("k = %d: (%d, %d) -- sending out local A matrix to (%d, %d).\n",
                      k, p, q, p, (qhat + j) % Q);
#endif
               Boardcast_dest = grid_to_global(P, Q, p, (qhat + j) % Q);
               CMMD_send_async(Boardcast_dest, Boardcast_tag, a[0], Buffer_Size,
                               send_handler, NULL); 
            }
         }
         else
         {
            /* boardcast the local matrix  */
#if 0
            printf("k = %d: (%d, %d) -- receiving local A matrix from (%d, %d).\n",
                   k, p, q, p, qhat);
#endif
            working_a_status = NOT_READY;
            Boardcast_src = grid_to_global(P, Q, p, qhat);

            Boardcast_mcb = CMMD_receive_async(Boardcast_src, Boardcast_tag,
                                               Buffer_a[0], Buffer_Size,
                                               receive_handler,
                                               &working_a_status); 
            if (free_temp_a == 1)
               free_temp_a = 2;
            else
               free_temp_a = 1;
         }
      }

      mult_local_matrix(m, n, working_a, working_b, c);
#if 0
      printf("(%d, %d) -- shifting local B matrix.\n", B -> p, B -> q);
#endif
   } while (k <= P);

#if 0
   if (p == q)
   {
      for (j = 0; j < m; j++)
         printf("%lf\n", c[j][j]);
   }
#endif
}

main()
{
   int      my_address;
   matrix   A;
   matrix   B;
   matrix   C;
   matrix   D;
   matrix   E;
   matrix   F;
   unsigned Seed;
   char     State[256];
   double   Total_Opt_per_node;
   
   my_address = CMMD_self_address();


   /* set independent IO mode */

   CMMD_fset_io_mode(stdout, CMMD_independent);
   CMMD_fset_io_mode(stderr, CMMD_independent);

   Seed = time(NULL) + CMMD_self_address() * 10000;
   initstate(Seed, State, 256);

   if (my_address < 16)
   {
      make_matrix(&A, MATRIX_SIZE, MATRIX_SIZE);
      clear_matrix(&A);
      fill_matrix(&A);

      make_matrix(&B, MATRIX_SIZE, MATRIX_SIZE);
      clear_matrix(&B);
      fill_matrix(&B);

      make_matrix(&C, MATRIX_SIZE, MATRIX_SIZE);
      clear_matrix(&C);

      make_matrix(&D, MATRIX_SIZE, MATRIX_SIZE);
      make_matrix(&E, MATRIX_SIZE, MATRIX_SIZE);
      make_matrix(&F, MATRIX_SIZE, MATRIX_SIZE);

      CMMD_sync_with_nodes();

      CMMD_node_timer_clear(0);
      CMMD_node_timer_start(0);

      matrix_mult(&A, &B, &C, &D, &E, &F);

      CMMD_sync_with_nodes();
      CMMD_node_timer_stop(0);

   }
   else
   {
      CMMD_sync_with_nodes();
      CMMD_node_timer_clear(0);
      CMMD_node_timer_start(0);

      CMMD_sync_with_nodes();
      CMMD_node_timer_stop(0);
   }

   Total_Opt_per_node = ((double) MATRIX_SIZE * MATRIX_SIZE * MATRIX_SIZE) / 16;

   if (my_address == 0)
   {
      printf("%d x %d\n", MATRIX_SIZE, MATRIX_SIZE);
      printf("Time=%lf [secs], flops per sec per node =%lf\n",
             CMMD_node_timer_busy(0),
             Total_Opt_per_node / CMMD_node_timer_busy(0));
   }

}

